import math
from PB3D import Vec4, Vec3, button_dict
from OpenGL.GLU import *

class Camera:
    """
    *** caution! This is not fully functional!***
    This is the class responsible for the camera in PB3D
    """
    def __init__(self, position=Vec3(0, 0, 0), rotation=Vec4(0, 0, 0, 1)):
        self.position = position
        self.rotation = rotation
        self.move_speed = 0.1
        self.mouse_sensitivity = 0.1

    def update(self, direction, speed):
        if direction == button_dict["w"]:
            self.move_forward(speed)
        elif direction == button_dict["s"]:
            self.move_backward(speed)

        self.set_view(self.position + self.rotation)

    def move_forward(self, speed):
        self.position += self.get_forward_vector() * speed

    def move_backward(self, speed):
        self.position -= self.get_forward_vector() * speed

    def rotate(self, dx, dy):
        dx *= self.mouse_sensitivity
        dy *= self.mouse_sensitivity

        rotation_quaternion = Vec4(
            -math.sin(math.radians(dx / 2)),
            math.sin(math.radians(dy / 2)),
            math.cos(math.radians(dx / 2)),
            math.cos(math.radians(dy / 2))
        )

        self.rotation = rotation_quaternion

        forward_vector = self.get_forward_vector()
        target_position = self.position + forward_vector
        self.set_view(target_position)

    def get_forward_vector(self):
        rotated_forward = self.rotation
        return Vec3(rotated_forward.x, rotated_forward.y, rotated_forward.z)

    def set_view(self, target_position):
        target = target_position
        direction = (self.position - target).normalize()

        up = Vec3(self.position.x, self.position.y + 1, self.position.z)
        right = (up.cross(direction)).normalize()

        camera_up = direction.cross(right)

        gluLookAt(self.position.x, self.position.y, self.position.z,
                  direction.x, direction.y, direction.z,
                  camera_up.x, camera_up.y, camera_up.z
                  )




from OpenGL.GL import *
from pygame.image import load, tostring

class Texture:
    def __init__(self, texture_path: str):
        self.texture_surface = load(texture_path)
        self.texture_data = tostring(self.texture_surface, "RGBA", 1)
        self.width, self.height = self.texture_surface.get_size()
        self.texture_id = glGenTextures(1)


from OpenGL.GLU import *
from pywavefront import Wavefront
from PB3D.math import Vec3, RGB
from PB3D.math.vector import BaseVec3
from OpenGL.GL import *

entities = []
GRAVITY = Vec3(9.8, 9.8, 9.8)

def mouse_pos(x, y):
    """
    This is a function that obtains the mouse click entity in PB3D.
    :param x:
    :param y:
    :return:
    """
    global selected_shape
    viewport = glGetIntegerv(GL_VIEWPORT)
    modelview = glGetDoublev(GL_MODELVIEW_MATRIX)
    projection = glGetDoublev(GL_PROJECTION_MATRIX)

    y = viewport[3] - y
    click_pos = gluUnProject(x, y, 0.0, modelview, projection, viewport)

    return click_pos



class PositionEntity:
    def __init__(self, position=BaseVec3(0, 0, 0)):
        self.position = position

    def __str__(self):
        return f"Position {self.position}"

class Scale:
    def __init__(self, num):
        self.scale = num

class MathEntity(PositionEntity, Scale):
    def __init__(self, position=BaseVec3(0, 0, 0), num=1):
        PositionEntity.__init__(self, position)
        Scale.__init__(self, num)


class Entity(MathEntity):
    """
    This is a class that displays the basic shapes of PB3D. This class supports color, position, etc. and also has an obj file loading function.
    """
    def __init__(self, file_path="cube", color=RGB(1, 1, 1), position=BaseVec3(0, 0, 0), scale=1, mass=Vec3(1, 1, 1)):
        super().__init__(
            position=position,
            num=scale
        )
        self.file_path = file_path
        self.color = color
        self.selected = False
        self.velocity = Vec3(0, 0, 0)
        self.mass = Vec3(self.scale, self.scale, self.scale)

        if file_path and file_path.endswith(".obj"):
            try:
                self.load_obj(file_path)
            except FileNotFoundError:
                self.obj_mesh = None
                self.vertices = [
                    (1 * self.scale + self.position.x, -1 * self.scale + self.position.y,
                     -1 * self.scale + self.position.z),
                    (1 * self.scale + self.position.x, 1 * self.scale + self.position.y,
                     -1 * self.scale + self.position.z),
                    (-1 * self.scale + self.position.x, 1 * self.scale + self.position.y,
                     -1 * self.scale + self.position.z),
                    (-1 * self.scale + self.position.x, -1 * self.scale + self.position.y,
                     -1 * self.scale + self.position.z),
                    (1 * self.scale + self.position.x, -1 * self.scale + self.position.y,
                     1 * self.scale + self.position.z),
                    (1 * self.scale + self.position.x, 1 * self.scale + self.position.y,
                     1 * self.scale + self.position.z),
                    (-1 * self.scale + self.position.x, -1 * self.scale + self.position.y,
                     1 * self.scale + self.position.z),
                    (-1 * self.scale + self.position.x, 1 * self.scale + self.position.y,
                     1 * self.scale + self.position.z)
                ]

        elif file_path == "cube":
            self.obj_mesh = None

        self.draw()
        entities.append(self)

    def load_obj(self, file_path):
        self.obj_mesh = Wavefront(file_path)

    def draw(self):
        glPushMatrix()
        glTranslatef(self.position.x, self.position.y, self.position.z)

        try:
            if self.obj_mesh:
                self.draw_obj()
            else:
                self.vertices = [
                    (1 * self.scale + self.position.x, -1 * self.scale + self.position.y,
                     -1 * self.scale + self.position.z),
                    (1 * self.scale + self.position.x, 1 * self.scale + self.position.y,
                     -1 * self.scale + self.position.z),
                    (-1 * self.scale + self.position.x, 1 * self.scale + self.position.y,
                     -1 * self.scale + self.position.z),
                    (-1 * self.scale + self.position.x, -1 * self.scale + self.position.y,
                     -1 * self.scale + self.position.z),
                    (1 * self.scale + self.position.x, -1 * self.scale + self.position.y,
                     1 * self.scale + self.position.z),
                    (1 * self.scale + self.position.x, 1 * self.scale + self.position.y,
                     1 * self.scale + self.position.z),
                    (-1 * self.scale + self.position.x, -1 * self.scale + self.position.y,
                     1 * self.scale + self.position.z),
                    (-1 * self.scale + self.position.x, 1 * self.scale + self.position.y,
                     1 * self.scale + self.position.z)
                ]
                self.draw_cube()
        except AttributeError:
            self.vertices = [
                (1 * self.scale + self.position.x, -1 * self.scale + self.position.y, -1 * self.scale + self.position.z),
                (1 * self.scale + self.position.x, 1 * self.scale + self.position.y, -1 * self.scale + self.position.z),
                (-1 * self.scale + self.position.x, 1 * self.scale + self.position.y, -1 * self.scale + self.position.z),
                (-1 * self.scale + self.position.x, -1 * self.scale + self.position.y, -1 * self.scale + self.position.z),
                (1 * self.scale + self.position.x, -1 * self.scale + self.position.y, 1 * self.scale + self.position.z),
                (1 * self.scale + self.position.x, 1 * self.scale + self.position.y, 1 * self.scale + self.position.z),
                (-1 * self.scale + self.position.x, -1 * self.scale + self.position.y, 1 * self.scale + self.position.z),
                (-1 * self.scale + self.position.x, 1 * self.scale + self.position.y, 1 * self.scale + self.position.z)
            ]
            self.draw_cube()

        glPopMatrix()

    def draw_cube(self):
        glPushMatrix()
        glTranslatef(self.position.x, self.position.y, self.position.z)

        if self.color:
            glColor3fv((self.color.red, self.color.green, self.color.blue))

        glBegin(GL_QUADS)
        for surface in ((0, 1, 2, 3), (3, 2, 7, 6), (4, 5, 1, 0), (1, 5, 7, 2), (4, 0, 3, 6)):
            for vertex_i in surface:
                try:
                    vertex = self.vertices[vertex_i]
                    glVertex3fv(vertex)
                except:
                    pass
        glEnd()

        glPopMatrix()

    def draw_obj(self):
        if self.selected:
            glColor3fv((1, 0, 0))
        elif self.color:
            glColor3fv((self.color.red, self.color.green, self.color.blue))

        glBegin(GL_TRIANGLES)
        for face in self.obj_mesh.mesh_list[0].faces:
            for vertex_i in face:
                vertex = self.obj_mesh.mesh_list[0].vertices[vertex_i]
                glVertex3fv(vertex)
        glEnd()

    def is_clicked(self, click_pos):
        if self.obj_mesh:
            min_x = min(v[0] for v in self.obj_mesh.mesh_list[0].vertices) + self.position.x
            max_x = max(v[0] for v in self.obj_mesh.mesh_list[0].vertices) + self.position.x
            min_y = min(v[1] for v in self.obj_mesh.mesh_list[0].vertices) + self.position.y
            max_y = max(v[1] for v in self.obj_mesh.mesh_list[0].vertices) + self.position.y
            min_z = min(v[2] for v in self.obj_mesh.mesh_list[0].vertices) + self.position.z
            max_z = max(v[2] for v in self.obj_mesh.mesh_list[0].vertices) + self.position.z
        else:
            min_x = min(v[0] for v in self.vertices) + self.position.x
            max_x = max(v[0] for v in self.vertices) + self.position.x
            min_y = min(v[1] for v in self.vertices) + self.position.y
            max_y = max(v[1] for v in self.vertices) + self.position.y
            min_z = min(v[2] for v in self.vertices) + self.position.z
            max_z = max(v[2] for v in self.vertices) + self.position.z

        return min_x <= click_pos[0] <= max_x and min_y <= click_pos[1] <= max_y and min_z <= click_pos[2] <= max_z

    def get_position(self):
        return self.position

    def set_position(self, position):
        if isinstance(position, tuple):
            self.position = Vec3(position[0], position[1], position[2])
        elif isinstance(position, Vec3):
            self.position = position

    def move(self):
        """
        This is the method responsible for the position movement function of the Shape.

        :return:
        """
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        displacement = self.velocity * 1/60

        self.position += displacement

        if self.file_path and self.file_path.endswith(".obj"):
            self.load_obj(self.file_path)
        elif self.file_path == "cube":
            self.vertices = [
                (1 * self.scale + self.position.x, -1 * self.scale + self.position.y, -1 * self.scale + self.position.z),
                (1 * self.scale + self.position.x, 1 * self.scale + self.position.y, -1 * self.scale + self.position.z),
                (-1 * self.scale + self.position.x, 1 * self.scale + self.position.y, -1 * self.scale + self.position.z),
                (-1 * self.scale + self.position.x, -1 * self.scale + self.position.y, -1 * self.scale + self.position.z),
                (1 * self.scale + self.position.x, -1 * self.scale + self.position.y, 1 * self.scale + self.position.z),
                (1 * self.scale + self.position.x, 1 * self.scale + self.position.y, 1 * self.scale + self.position.z),
                (-1 * self.scale + self.position.x, -1 * self.scale + self.position.y, 1 * self.scale + self.position.z),
                (-1 * self.scale + self.position.x, 1 * self.scale + self.position.y, 1 * self.scale + self.position.z)
            ]
            self.draw_cube()

    def gravity(self):
        for entity in entities:
            if isinstance(entity, Entity) and entity != self:
                displacement = entity.position - self.position
                distance_squared = displacement.dot_product(displacement)
                force_direction = displacement.normalize()
                force_magnitude = GRAVITY * ((self.mass * entity.mass) / distance_squared)
                force = force_direction * force_magnitude
                acceleration = force / self.mass
                self.velocity += acceleration * (1 / 60)
        self.move()


def delete(entity: Entity):
    entity.position = Vec3(0, 0, 0)
    entity.scale = 0
    entity.vertices = []
    del entity

class Shape2d:
    """
    *** caution! This does not add any new features in version 0.0.3. ***
    This is a class that represents 2d entities in PB3D.
    """
    def __init__(self, file_path="square", color=None, position=(0, 0)):
        self.file_path = file_path
        self.color = color
        self.position = position
        self.selected = False

        glPushMatrix()
        glTranslatef(self.position[0], self.position[1], 0.0)

        if self.selected:
            glColor3fv((1, 0, 0))
        elif self.color:
            glColor3fv(self.color)

        glBegin(GL_QUADS)
        glVertex2f(0, 0)
        glVertex2f(50, 0)
        glVertex2f(50, 50)
        glVertex2f(0, 50)
        glEnd()

        glPopMatrix()

    def is_clicked(self, click_pos):
        min_x = self.position[0]
        max_x = self.position[0] + 50
        min_y = self.position[1]
        max_y = self.position[1] + 50

        return min_x <= click_pos[0] <= max_x and min_y <= click_pos[1] <= max_y

    def get_position(self):
        return self.position

    def set_position(self, position):
        self.position = position

    def move(self, x, y):
        self.position = (self.position[0] + x, self.position[1] + y)


class Light:
    """
    *** caution! This is not a complete feature yet. ***
    This is a class that represents light in PB3D.
    """
    def __init__(self, position=(0, 0, 0), ambient=1, diffuse=1, specular=1):
        self.position = position
        self.ambient = [ambient, ambient, ambient, 1.0]
        self.diffuse = [diffuse, diffuse, diffuse, 1.0]
        self.specular = [specular, specular, specular, 1.0]

        glEnable(GL_LIGHTING)
        glEnable(GL_LIGHT0)
        self.update()

    def update(self):
        glLightfv(GL_LIGHT0, GL_POSITION, self.position)
        glLightfv(GL_LIGHT0, GL_AMBIENT, self.ambient)
        glLightfv(GL_LIGHT0, GL_DIFFUSE, self.diffuse)
        glLightfv(GL_LIGHT0, GL_SPECULAR, self.specular)

    def set_position(self, position):
        self.position = position
        self.update()

    def set_ambient(self, ambient):
        self.ambient = [ambient, ambient, ambient, 1.0]
        self.update()

    def set_diffuse(self, diffuse):
        self.diffuse = [diffuse, diffuse, diffuse, 1.0]
        self.update()

    def set_specular(self, specular):
        self.specular = [specular, specular, specular, 1.0]
        self.update()



from pygame.event import get

button_dict = {i: ord(i) for i in "abcdefghijklmnopqrstuvwxyz0123456789\n\t\b\\\'\""}
button_dict["down"] = 768
button_dict["mouse"] = 1025

def event_get():
    return get()


from .window import init, init_2d, loop, Event, turn
from .entity import Entity, mouse_pos, entities, Shape2d, Light, delete
from .event import button_dict, event_get
from .math import Vec3, HSV, RGB, Vec4, i, j, k
from .texture import Texture
from .camera import Camera



import colorsys

class RGB:
    """
    This is a class that implements the RBG system, one of PB3D's color systems. Colors are easy to use because they take three arguments with a maximum value of 255.
    """
    def __init__(self, red, green, blue):
        self.red = red/255
        self.green = green/255
        self.blue = blue/255

    def hsv(self):
        rgb = colorsys.rgb_to_hsv(self.red, self.green, self.blue)
        return HSV(rgb[0]*360, rgb[1], rgb[2])

class HSV:
    """
    This is the HSV system, one of PB3D's color systems. Takes color, saturation, and brightness parameters
    """
    def __init__(self, hue, saturation, value):
        self.hue = hue
        self.saturation = saturation
        self.value = value


    def rgb(self):
        rgb = colorsys.hsv_to_rgb(self.hue / 360, self.saturation, self.value)
        return RGB(rgb[0] * 255, rgb[1] * 255, rgb[2] * 255)


import sys

import pygame
from pygame.locals import *
from OpenGL.GL import *
from OpenGL.GLU import gluPerspective
from PB3D.math import Vec4, RGB
from PB3D.entity import entities, delete
from OpenGL.GL import glEnable, GL_DEPTH_TEST
from numba import jit, NumbaWarning
import warnings

warnings.filterwarnings("ignore", category=NumbaWarning)


def change_color(color: RGB):
    glClearColor(color.red, color.green, color.blue, 1)

class Window:
    def __init__(self, size: tuple[int, int], color: RGB):
        self.width = size[0]
        self.height = size[1]
        self.color = color


@jit
def init(size: tuple[int, int], color: RGB):
    """
    This is a function that initializes the 3d mode of PB3D. Here you can adjust the color and size.

    :param size:
    :param color:
    :return:
    """
    pygame.init()
    pygame.display.set_mode(size, DOUBLEBUF | OPENGL)

    glEnable(GL_DEPTH_TEST)

    gluPerspective(45, (size[0] / size[1]), 0.1, 50.0)
    glTranslatef(0.0, 0.0, -5)
    glClearColor(color.red, color.green, color.blue, 1)

    print("\nOpenGL Version:", glGetString(GL_VERSION))
    print("OpenGL Vendor:", glGetString(GL_VENDOR))
    print("OpenGL Renderer:", glGetString(GL_RENDERER))
    print("\nVersion 1.0.1")

    return Window(size, color)

@jit
def init_2d(size: tuple[int, int]):
    """
    This is a function that initializes the 2d mode of PB3D. Here you can adjust the color and size.
    :param size:
    :return:
    """
    pygame.init()
    pygame.display.set_mode(size, DOUBLEBUF | OPENGL)
    glOrtho(0, size[0], size[1], 0, -1, 1)

    print("\nOpenGL Version:", glGetString(GL_VERSION))
    print("OpenGL Vendor:", glGetString(GL_VENDOR))
    print("OpenGL Renderer:", glGetString(GL_RENDERER))
    print("\nVersion 0.0.3")

@jit
def update():
    pygame.display.flip()

@jit
def clean():
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

@jit
def turn(vec: Vec4):
    """
    This is the function responsible for adjusting the field of view in PB3D
    :param vec
    :return:
    """
    glRotatef(vec.w, vec.x, vec.y, vec.z)

@jit
def loop(func1=None, func2=None, func3=None):
    """
    This is a function that manages loops in PB3D. Basically, you can install a model in func1 and use the keyboard in func2.
    :param func1:
    :param func2:
    :return:
    """
    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                quit()
            if event.type == pygame.KEYDOWN:
                if func2 != None:
                    func2(event)

            if event.type == pygame.MOUSEMOTION:
                dx, dy = event.rel
                if func3 != None:
                    func3(dx, dy)

        clean()
        if func1 != None:
            func1()

        for entity in entities:
            entity.draw()

            if entity.position.size > round(sys.maxsize / 8):
                delete(entity)

        update()

Event = pygame.event.Event


import math

class BaseVec3:
    """
    This is the base class for classes expressing coordinates in PB3D and supports four arithmetic operations on coordinates.
    """
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z
        self.size = (self.x ** 2 + self.y ** 2 + self.z ** 2) ** 0.5

    def __add__(self, other):
        return BaseVec3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        return BaseVec3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, scalar):
        return BaseVec3(self.x * scalar, self.y * scalar, self.z * scalar)

    def __truediv__(self, scalar):
        if scalar != 0:
            return BaseVec3(self.x / scalar, self.y / scalar, self.z / scalar)
        else:
            raise ValueError("Cannot divide by zero.")

    def __str__(self):
        return f"Vec3(x={self.x}, y={self.y}, z={self.z})"

class Vec3(BaseVec3):
    """
    This is a class that represents coordinates in PB3D and provides several mathematical functions.
    """
    def __init__(self, x, y, z):
        super().__init__(x, y, z)
        if self.size > 0:
            self.unit = BaseVec3((self.x / self.size), (self.y / self.size), (self.z / self.size))
            self.direction = BaseVec3(
                math.acos(self.unit.x),
                math.acos(self.unit.y),
                math.acos(self.unit.z)
            )
        else:
            self.unit = BaseVec3(0, 0, 0)
            self.direction = BaseVec3(0, 0, 0)

    def distance(self, other):
        x = abs(self.x - other.x)
        y = abs(self.y - other.y)
        z = abs(self.z - other.z)

        return math.sqrt(x ** 2 + y ** 2 + z ** 2)

    def magnitude(self):
        return math.sqrt(self.x**2 + self.y**2 + self.z**2)

    def normalize(self):
        mag = self.magnitude()
        if mag != 0:
            return Vec3(self.x / mag, self.y / mag, self.z / mag)
        else:
            return Vec3(0, 0, 0)

    def dot_product(self, other):
        if isinstance(other, Vec3):
            return Vec3(self.x * other.x, self.y * other.y, self.z * other.z)
        elif isinstance(other, (int, float)):
            return Vec3(self.x * other, self.y * other, self.z * other)
        else:
            raise TypeError("Unsupported operand type for dot product")

    def cross(self, other):
        return Vec3(
            self.y * other.z - self.z * other.y,
            self.z * other.x - self.x * other.z,
            self.x * other.y - self.y * other.x
        )

    def angle(self, other, in_degrees=False):
        dot_product = self.dot_product(other)
        mag_self = self.magnitude()
        mag_other = other.magnitude()

        if mag_self != 0 and mag_other != 0:
            cos_theta = dot_product / (mag_self * mag_other)
            theta_rad = math.acos(max(min(cos_theta, 1.0), -1.0))

            if in_degrees:
                return math.degrees(theta_rad)
            else:
                return theta_rad
        else:
            raise ValueError("Cannot calculate angle for zero vectors.")

    def __add__(self, other):
        return Vec3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other):
        return Vec3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, scalar):
        if isinstance(scalar, (int, float)):
            return Vec3(self.x * scalar, self.y * scalar, self.z * scalar)
        else:
            return Vec3(self.x * scalar.x, self.y * scalar.y, self.z * scalar.z)

    def __truediv__(self, scalar):
        if isinstance(scalar, (int, float)):
            if scalar != 0:
                return Vec3(self.x / scalar, self.y / scalar, self.z / scalar)
            else:
                return Vec3(0, 0, 0)
        elif isinstance(scalar, Vec3):
            x = scalar.x if scalar.x != 0 else 1
            y = scalar.y if scalar.y != 0 else 1
            z = scalar.z if scalar.z != 0 else 1
            return Vec3(self.x / x, self.y / y, self.z / z)
        else:
            raise TypeError("Unsupported operand type for division.")


class Vec4:
    def __init__(self, a, b, c, d):
        self.x = a
        self.y = b
        self.z = c
        self.w = d

    def __repr__(self):
        return f"Vec4({self.x}, {self.y}, {self.z}, {self.w})"

    def __add__(self, other):
        if isinstance(other, Vec4):
            return Vec4(self.x + other.x, self.y + other.y, self.z + other.z, self.w + other.w)
        elif isinstance(other, (int, float)):
            return Vec4(self.x + other, self.y + other, self.z + other, self.w + other)
        else:
            raise TypeError("Unsupported operand type for +: Vec4 and {}".format(type(other)))

    def __sub__(self, other):
        if isinstance(other, Vec4):
            return Vec4(self.x - other.x, self.y - other.y, self.z - other.z, self.w - other.w)
        elif isinstance(other, (int, float)):
            return Vec4(self.x - other, self.y - other, self.z - other, self.w - other)
        else:
            raise TypeError("Unsupported operand type for -: Vec4 and {}".format(type(other)))

    def __mul__(self, other):
        if isinstance(other, Vec4):
            a1, b1, c1, d1 = self.x, self.y, self.z, self.w
            a2, b2, c2, d2 = other.x, other.y, other.z, other.w

            a = a1 * a2 - b1 * b2 - c1 * c2 - d1 * d2
            b = a1 * b2 + b1 * a2 + c1 * d2 - d1 * c2
            c = a1 * c2 - b1 * d2 + c1 * a2 + d1 * b2
            d = a1 * d2 + b1 * c2 - c1 * b2 + d1 * a2

            return Vec4(a, b, c, d)
        elif isinstance(other, (int, float)):
            return Vec4(self.x * other, self.y * other, self.z * other, self.w * other)
        else:
            raise TypeError("Unsupported operand type for *: Vec4 and {}".format(type(other)))

    def conjugate(self):
        return Vec4(self.x, -self.y, -self.z, -self.w)

    def normalize(self):
        norm = (self.x ** 2 + self.y ** 2 + self.z ** 2 + self.w ** 2) ** 0.5
        if norm == 0:
            return Vec4(0, 0, 0, 0)
        else:
            return Vec4(self.x / norm, self.y / norm, self.z / norm, self.w / norm)

    def __truediv__(self, other):
        if isinstance(other, Vec4):
            try:
                x = self.x / other.x
            except ZeroDivisionError:
                x = 0

            try:
                y = self.y / other.y
            except ZeroDivisionError:
                y = 0

            try:
                z = self.z / other.z
            except ZeroDivisionError:
                z = 0

            try:
                w = self.w / other.w
            except ZeroDivisionError:
                w = 0

            return Vec4(x, y, z, w)
        elif isinstance(other, (int, float)):
            if other != 0:
                return Vec4(self.x / other, self.y / other, self.z / other, self.w / other)
            else:
                raise ZeroDivisionError("division by zero")
        else:
            raise TypeError("Unsupported operand type for /: Vec4 and {}".format(type(other)))

    def rotate_vector(self, v):
        q_conj = Vec4(self.w, -self.x, -self.y, -self.z)
        qv = Vec4(0, v.x, v.y, v.z)
        rotated_v = (self * qv * q_conj).xyz()
        return Vec4(rotated_v.x, rotated_v.y, rotated_v.z, 0)

    def xyz(self):
        return Vec3(self.x, self.y, self.z)

i = Vec4(1, 0, 0, 0)
j = Vec4(0, 1, 0, 0)
k = Vec4(0, 0, 1, 0)
r = Vec4(0, 0, 0, 1)



import setuptools

setuptools.setup(
    name="PB3D",
    version="1.1.0a1",
    author="InvincibleSiwoo",
    author_email="siukk@gmail.com",
    description="A small engine",
    long_description="""
# PB3D

at las! PB3D has been officially released as version 1.0.0!
Changed: Added camera direction
                        """,
    long_description_content_type="text/markdown",
    url="https://github.com/InvincibleSiwoo/engine/tree/master",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.11',
)
